# -*- coding: utf-8 -*-
import tkinter as tk
import customtkinter
from tkinter import filedialog as fd
from tkinter.messagebox import showinfo
from tkinter import messagebox
from tkinter import ttk
from functools import partial
import numpy as np
import MistrasDTA
import pandas as pd
import os
from datetime import datetime
from datetime import date
import gc

customtkinter.set_appearance_mode("dark")  # Modes: system (default), light, dark
customtkinter.set_default_color_theme("blue")  # Themes: blue (default), dark-blue, green
    
root = customtkinter.CTk()  
root.tk.call('tk', 'scaling', 1.4)
root.geometry('510x355')     
root.resizable(width=False, height=False)
root.title('DTA extract')  

#-- Functions --#
def rmfield( a, *fieldnames_to_remove ):
    return a[ [ name for name in a.dtype.names if name not in fieldnames_to_remove ] ]

def loadfile(label_loadfile, label_loadpath):
    filetypes=(("DTA", ".dta"),)
    filename = fd.askopenfilename(
        title='Open a file',
        initialdir='/',
        filetypes=filetypes)
    if filename is None:
        return
    path = os.path.dirname(filename)
    label_loadfile.config(text="%s" % filename)
    label_loadpath.config(text="%s" % path)
    return 

def extract():
    if not os.path.exists(os.path.join(label_loadpath.cget("text"),'eventwaveforms\\')):
        os.makedirs(os.path.join(label_loadpath.cget("text"),'eventwaveforms\\'))

    if not os.path.exists(label_loadDTA.cget("text")):
        return
    
    showinfo(
        title='Selected File',
        message="Please be patient! A new message will pop up when all waveform files have been written to disk.")
    
    # Read the binary file
    rec, wfm = MistrasDTA.read_bin(label_loadDTA.cget("text"))

    #Convert to Dataframe (Pandas)
    wfm_array = wfm.view(np.recarray)
    wfm_df_full = pd.DataFrame(wfm_array)
    rec_array = rec.view(np.recarray)
    rec_df_full = pd.DataFrame(rec_array)
    
    #Remove Channel of choice
    if int(float(Ch.get())) == 0:
        wfm_df = wfm_df_full
        rec_df = rec_df_full
    else:
        wfm_df = wfm_df_full.drop(wfm_df_full[wfm_df_full.CH == int(Ch.get())].index)
        rec_df = rec_df_full.drop(rec_df_full[rec_df_full.CH == int(Ch.get())].index)
    
    #Count unique number of channels
    CountCh = rec_df.CH.unique().size
    
    #Set mask with precision on event identification to x number of decimals in time
    wfm_round = np.around(wfm_df.iloc[:,0], decimals=int(prec.get()))
    wfm_join = pd.merge(wfm_round, wfm_df, left_index=True, right_index=True, how="outer")
    rec_round = np.around(rec_df.iloc[:,0], decimals=int(prec.get()))
    rec_join = pd.merge(rec_round, rec_df, left_index=True, right_index=True, how="outer")

    #Extract events on x channels
    EventsCha = wfm_join.groupby('SSSSSSSS.mmmuuun_x').filter(lambda x: len(x) == CountCh)
    EventsCh = EventsCha.sort_values(by=['SSSSSSSS.mmmuuun_x', 'CH'])
    AEcharsCha = rec_join.groupby('SSSSSSSS.mmmuuun_x').filter(lambda x: len(x) == CountCh)
    AEcharsCh = AEcharsCha.sort_values(by=['SSSSSSSS.mmmuuun_x','CH'])
    
    #Round to 7th decimal place for time in AEcharsCh
    AEcharsCh = AEcharsCh.round({'SSSSSSSS.mmmuuun_y': 7})
    
    #Add fake index column
    AEcharsCh.insert(0, "Index", AEcharsCh['SSSSSSSS.mmmuuun_x'] / AEcharsCh['SSSSSSSS.mmmuuun_x'], True) 
    AEcharsCh = AEcharsCh.astype({'Index':'int'})
    
    #Convert back to Record array
    AECh = EventsCh.to_records()
    
    if len(AECh) == 0:
        tk.messagebox.showwarning(title="No events!", message="Based on the chosen time precision, no events could be found. Try a lower value. If applicable, you may consider excluding the channel with the cBN anvil due to arrival time differences with the other WC anvils.")
        return
    
    AECh = np.sort(AECh, order=['SSSSSSSS.mmmuuun_x', 'CH'])
    AECh_chars = AEcharsCh.to_records()
    AECh_chars = np.sort(AECh_chars, order=['SSSSSSSS.mmmuuun_x', 'CH'])
    AECh_chars = rmfield(AECh_chars, 'index')
    AECh_chars = rmfield(AECh_chars, 'SSSSSSSS.mmmuuun_x')
        
    #Clean up memory
    del(rec)
    del(wfm)
    del(wfm_array)
    del(wfm_df)
    del(rec_array)
    del(rec_df)
    del(wfm_round)
    del(wfm_join)
    del(rec_round)
    del(rec_join)
    gc.collect()
    
    now = datetime.now()
    today = date.today()
    dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
    d1 = today.strftime("%A, %B %d, %Y") 
    t1 = now.strftime("%H:%M:%S")
    
    #Save waveforms as csv files with filenames formatted as: Exp#_Channel_Event#_Time.csv
    for i in range(len(AECh)):
        t, V = MistrasDTA.get_waveform_data(AECh[i])
        channel = EventsCh.iloc[i,2]
        
        if i % CountCh == 0:
            event = 1 + int(i / CountCh)
        
        time = round(EventsCh.iloc[i,1] * 1E7)
       
        fname = os.path.join(label_loadpath.cget("text"),'eventwaveforms\\',f'{exp.get()}_{channel}_{event}_{time}.csv')
        np.savetxt(fname, V, fmt='%.14f', comments='', header=f'SOURCE FILE NAME: {label_loadDTA.cget("text")}\nDATE: {d1}\nTIME: {t1}\nSAMPLE INTERVAL (Seconds):   0.0000000250\nSIGNAL UNITS: volts\nTIME UNITS: Seconds\nDATA TYPE: WAVEFORM\nNUMBER OF DATA POINTS PER WAVEFORM: 15360\nCHANNEL NUMBER: {channel}\nHIT NUMBER: {i + 1}\nTIME OF TEST: {time / 1E7}\n')
    
    #Save AE characteristics of each hit of each event in txt file:
    fname = os.path.join(label_loadpath.cget("text"),'eventwaveforms\\',f'{exp.get()}_AEchars.txt')
    np.savetxt(fname, AECh_chars, fmt='%s', comments='', header=f'{label_loadDTA.cget("text")}\nPCI2 (r) Location Version\n\nVersion P61B\n\n\n{dt_string}\n\nID SSSSSSSS.mmmuuun CH RISE COUN ENER DURATION AMP ASL THR A-FRQ ABS-ENERGY FRQ-C P-FRQ TIMESTAMP')
    
    messagebox.showinfo("Complete", "All files have been written to disk!")
    
    return

def extractall():
    if not os.path.exists(os.path.join(label_loadpath.cget("text"),'allwaveforms\\')):
        os.makedirs(os.path.join(label_loadpath.cget("text"),'allwaveforms\\'))

    if not os.path.exists(label_loadDTA.cget("text")):
        return
    
    showinfo(
        title='Selected File',
        message="Please be patient! A new message will pop up when all waveform files have been written to disk.")
    
    # Read the binary file
    rec, wfm = MistrasDTA.read_bin(label_loadDTA.cget("text"))

    #Convert to Dataframe (Pandas)
    rec_array =rec.view(np.recarray)
    rec_df = pd.DataFrame(rec_array)

    now = datetime.now()
    today = date.today()
    dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
    d1 = today.strftime("%A, %B %d, %Y") 
    t1 = now.strftime("%H:%M:%S")
    
    #Save waveforms as csv files with filenames formatted as: Exp#_Channel_Event#_Time.csv
    for i in range(len(wfm)):
        t, V = MistrasDTA.get_waveform_data(wfm[i])
        channel = rec_df.iloc[i,1]
        
        time = round(rec_df.iloc[i,0] * 1E7)
       
        fname = os.path.join(label_loadpath.cget("text"),'allwaveforms\\',f'{exp.get()}_{channel}_{time}.csv')
        np.savetxt(fname, V, fmt='%.14f', comments='', header=f'SOURCE FILE NAME: {label_loadDTA.cget("text")}\nDATE: {d1}\nTIME: {t1}\nSAMPLE INTERVAL (Seconds):   0.0000000250\nSIGNAL UNITS: volts\nTIME UNITS: Seconds\nDATA TYPE: WAVEFORM\nNUMBER OF DATA POINTS PER WAVEFORM: 15360\nCHANNEL NUMBER: {channel}\nHIT NUMBER: {i + 1}\nTIME OF TEST: {time / 1E7}\n')
    
    #Save AE characteristics of each hit of each event in txt file:
    fname = os.path.join(label_loadpath.cget("text"),'allwaveforms\\',f'{exp.get()}_AEchars.txt')
    np.savetxt(fname, rec, fmt='%s', comments='', header=f'{label_loadDTA.cget("text")}\nPCI2 (r) Location Version\n\nVersion P61B\n\n\n{dt_string}\n\nID SSSSSSSS.mmmuuun CH RISE COUN ENER DURATION AMP ASL THR A-FRQ ABS-ENERGY FRQ-C P-FRQ TIMESTAMP')
    
    messagebox.showinfo("Complete", "All files have been written to disk!")
    
    return

#-- GUI ROW BY ROW --# 
#Row 0
label_RunNumber = tk.Label(root, text="Enter experiment run number")
label_RunNumber.grid(row=0, column=0, sticky=tk.W, columnspan=6, pady=(10,0), padx=(10, 0))
label_RunNumber.config(bg="#212325", fg="#FFFFFF", font='Helvetica 12 bold')

#Row 1
exp = tk.StringVar()
entry_exp = customtkinter.CTkEntry(master=root, textvariable=exp, border_width = 2, width = 100, justify='center')
entry_exp.grid(row=1, column=0, sticky=tk.W, padx=(15, 5), pady=(5,5)) 
entry_exp.insert(0, 'BT___')

#Row 2 Separator
separator = ttk.Separator(root, orient='horizontal')
separator.grid(row=2, column=0, columnspan=8, sticky="ew", padx=(15, 5), pady=(10,2))

#Row 3
label_Channel = tk.Label(root, text="1. Extract waveforms and AE data for events only")
label_Channel.grid(row=3, column=0, sticky=tk.W, columnspan=6, pady=(5,0), padx=(10, 0))
label_Channel.config(bg="#212325", fg="#FFFFFF", font='Helvetica 12 bold')

#Row 4
Ch = tk.StringVar()
entry_Ch = customtkinter.CTkEntry(master=root, textvariable=Ch, border_width = 2, width = 100, justify='center')
entry_Ch.grid(row=4, column=0, sticky=tk.W, padx=(15, 5), pady=(5,5)) 
entry_Ch.insert(0, '0')

label_Channel = tk.Label(root, text="Enter Ch # to remove ('0' is none)")
label_Channel.grid(row=4, column=1, sticky=tk.W, columnspan=6, pady=(5,0), padx=(10, 0))
label_Channel.config(bg="#212325", fg="#FFFFFF", font='Helvetica 12')

#Row 5
prec = tk.StringVar()
entry_prec = customtkinter.CTkEntry(master=root, textvariable=prec, border_width = 2, width = 100, justify='center')
entry_prec.grid(row=5, column=0, sticky=tk.W, padx=(15, 5), pady=(5,5)) 
entry_prec.insert(0, '6')

label_Precision = tk.Label(root, text="Arrival time precision (# of decimal places, max 7)")
label_Precision.grid(row=5, column=1, sticky=tk.W, columnspan=6, pady=(5,0), padx=(10, 0))
label_Precision.config(bg="#212325", fg="#FFFFFF", font='Helvetica 12')

#Row 6
label_loadDTA = tk.Label(root, text="path") #redirect for another function
label_loadpath = tk.Label(root, text="path") #redirect for another function
#-- Button --#
loadDTA = partial(loadfile, label_loadDTA, label_loadpath)
buttonCal = customtkinter.CTkButton(root, text="Load DTA file", width = 100, command=lambda:[loadDTA(),extract()])
buttonCal.grid(row=6, column=0, sticky=tk.W, padx=(15, 5), pady=(2,2))
label_Import = tk.Label(root, text="Extract event data")
label_Import.grid(row=6, column=1, sticky=tk.W, columnspan=6, pady=(5,0), padx=(10, 0))
label_Import.config(bg="#212325", fg="#FFFFFF", font='Helvetica 12')

#Row 7
separator = ttk.Separator(root, orient='horizontal')
separator.grid(row=7, column=0, columnspan=8, sticky="ew", padx=(15, 5), pady=(10,2))

#Row 8
label_Import = tk.Label(root, text="2. Extract waveforms and AE data for all hits")
label_Import.grid(row=8, column=0, sticky=tk.W, columnspan=6, pady=(5,0), padx=(10, 0))
label_Import.config(bg="#212325", fg="#FFFFFF", font='Helvetica 12 bold')

#Row 9
#-- Button --#
loadDTA = partial(loadfile, label_loadDTA, label_loadpath)
buttonCal = customtkinter.CTkButton(root, text="Load DTA file", width = 100, command=lambda:[loadDTA(),extractall()])
buttonCal.grid(row=9, column=0, sticky=tk.W, padx=(15, 5), pady=(2,2))
label_Import = tk.Label(root, text="Extract all data")
label_Import.grid(row=9, column=1, sticky=tk.W, columnspan=6, pady=(5,0), padx=(10, 0))
label_Import.config(bg="#212325", fg="#FFFFFF", font='Helvetica 12')

#Row 10
label_Note = tk.Label(root, text="Written by R. Farla at P61B      20.12.2023      v1.3")
label_Note.grid(row=10, column=1, sticky=tk.E, columnspan=6, pady=(25,0))
label_Note.config(bg="#212325", fg="#FFFFFF", font='Helvetica 9 bold')

root.mainloop()